#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>

#include "f81504.h"

/*---------------------------------------------------------------------------------*/
#define F81504_CONFIG_SYSFS 	   "/sys/class/tty/%s/device/config"
#define F81504_CONFIG_SYSFS_IOPORT "/sys/class/tty/%s/port"
#define F81504_PCI_SYSFS 		   "/sys/class/tty/%s/device"

/*---------------------------------------------------------------------------------*/
enum error
{
	error_no_error = 0,

	error_parameter,
	error_open_failed,
	error_not_f81504,
	error_change_mode,
	error_read_id,
	error_read_io,
	error_read_pci,
	error_write_pci,
	error_io_not_found,
};

/*---------------------------------------------------------------------------------*/
static char *transceiver_type_str[4] =
{
	"F81439",
	"F81435",
	"D81433",
	"Unknown"
};

/*=================================================================================*/
/* Fintek */
/*---------------------------------------------------------------------------------*/
static int read_ioport(char *tty)
{
	char buf[64];
	int  len;
	int  fd;
	int  io_addr;

	sprintf(buf, F81504_CONFIG_SYSFS_IOPORT, tty);
	//printf("Target:%s\n", buf);

	fd = open(buf, O_RDONLY);
	if(fd < 0)
	{
		// fprintf(stderr, "open error %d\n", __LINE__);
		// perror("");
		return -1;
	}

	memset(buf, 0, sizeof(buf));
	len = read(fd, buf, sizeof(buf));
	if(!len)
		return -1;

	io_addr = strtol(buf, NULL, 16);
	close(fd);

	return io_addr;

} /* END: read_ioport() */

/*---------------------------------------------------------------------------------*/
static int read_pci_byte(char *name, int addr, unsigned char *data)
{
	FILE *file;
	int   len;
	char  m_pci_path[64];

	sprintf(m_pci_path, F81504_CONFIG_SYSFS, name);

	file = fopen(m_pci_path, "r");
	if(!file)
		return -1;

	fseek(file, addr, SEEK_SET);
	len = fread(data, 1, 1, file);
	if(len != 1)
	{
		fclose(file);
		return -1;
	}

	fclose(file);

	return 0;

} /* END: read_pci_byte() */

/*---------------------------------------------------------------------------------*/
static int get_tty_idx(char *name, int io_addr)
{
	unsigned char buf[2];
	unsigned int  i, j;
	int 		  read_io;

	for (i=0; i<12; ++i)
	{
		memset(buf, 0, sizeof(buf));

		for(j=0; j<2; ++j)
		{
			read_pci_byte(name, 0x40 + 0x8 * i + 4 + 0, &buf[0]);
			read_pci_byte(name, 0x40 + 0x8 * i + 4 + 1, &buf[1]);
		}

		read_io = (buf[1] << 8) | buf[0];

		if(io_addr == read_io)
			break;
	}

	if (i >= 12)
		return -1;

	return i;

} /* END: get_tty_idx() */

/*---------------------------------------------------------------------------------*/
static int get_transceiver_type(char *tty, int idx, int *type)
{
	char  buf[128];
	FILE *file;
	int   i, data[4];

	if(idx >= 4)
		return -1;

	sprintf(buf, F81504_PCI_SYSFS "/transceiver", tty);

	file = fopen(buf, "r");
	if(!file)
	{
		// printf("open failed: %s\n", buf);
		return -1;
	}

	for(i=0; i<4; ++i)
	{
		fscanf(file, "%d\n", &data[i]);
		//printf("%s: %d %d\n", __func__, i, data[i]);

		if (i == idx)
			*type = data[i];
	}

	fclose(file);
	return 0;

} /* END: get_transceiver_type() */

/*---------------------------------------------------------------------------------*/
static int get_current_mode(char *tty, int idx, int *mode)
{
	char  buf[128];
	FILE *file;
	int   i, data[4];

	if(idx >= 4)
		return -1;

	sprintf(buf, F81504_PCI_SYSFS "/gpio_ctrl", tty);

	file = fopen(buf, "r");
	if(!file)
	{
		// printf("open failed: %s\n", buf);
		return -1;
	}

	for(i=0; i<4; ++i)
	{
		fscanf(file, "%d\n", &data[i]);
		//printf("%s: %d %d\n", __func__, i, data[i]);

		if(i == idx)
			*mode = data[i];
	}

	return 0;

} /* END: get_current_mode() */

/*---------------------------------------------------------------------------------*/
static int change_mode(int fd, int mode, int io_addr)
{
	unsigned char buf[4];
	int 		  read_io;
	int 		  len;
	int 		  i;

	for(i=0; i<12; ++i)
	{
		memset(buf, 0, sizeof(buf));
		lseek(fd, 0x40 + 0x8 * i + 4, SEEK_SET);

		len = read(fd, buf, 2);
		if(len != 2)
			return error_read_pci;

		read_io = (buf[1] << 8) | buf[0];

		if(io_addr == read_io)
			break;
	}

	if(i >= 12)
		return error_io_not_found;

	memset(buf, 0, sizeof(buf));
	lseek(fd, 0x40 + 0x8 * i + 7, SEEK_SET);

	len = read(fd, buf, 1);
	if(len != 1)
		return error_read_pci;

	switch(mode)
	{
		case 1:
			buf[0] = 0x01;
			break;
		case 2:
			buf[0] = 0x31;
			break;
		case 3:
			buf[0] = 0x11;
			break;
		default:
			return error_parameter;
	}

	lseek(fd, 0x40 + 0x8 * i + 7, SEEK_SET);
	len = write(fd, buf, 1);
	if(len != 1)
		return error_write_pci;
	
	return error_no_error;

} /* END: change_mode() */

/*---------------------------------------------------------------------------------*/
static int change_gpio(char *tty, int idx, unsigned int mode)
{
	int	  ret  = 0;
	char  buf[128];
	FILE *file = NULL;
	int   len;

	if(idx >= 4)
		return -1;

	sprintf(buf, F81504_PCI_SYSFS "/gpio_ctrl", tty);

	file = fopen(buf, "w");
	if(!file)
	{
		ret = 1;
		goto EXIT;
	}

	sprintf(buf, "%d %d\n", idx, mode);

	len = fprintf(file, buf, strlen(buf));
	if(len <= 0)
	{
		// printf("set gpio failed: %d\n", len);
		ret = 2;
		goto EXIT;
	}

EXIT:

	if(file)
		fclose(file);

	return ret;

} /* END: change_gpio() */

/*---------------------------------------------------------------------------------*/
int is_f81504(char *tty_port)
{
	int 		  len;
	unsigned char buf[5];

	memset(buf, 0, sizeof(buf));

	for(len=0; len<4; ++len)
	{
		read_pci_byte(tty_port, len, &buf[len]);
		//printf("%x", buf[len]);
	}

	//printf("\n");

	if((buf[0] != 0x29) || (buf[1] != 0x1c) || (buf[3] != 0x11))
		return -1;

	switch(buf[2])
	{
		case 0x04:
		case 0x08:
		case 0x12:
			return 0;
	}

	return -1;

} /* END: is_f81504() */

/*=================================================================================*/
/* Innodisk */
/*---------------------------------------------------------------------------------*/
static char *find_mode_str_by_cmd_gpio(int cmd_gpio)
{
	int				  i;
	MODE_F81439_INFO *info = NULL;

	for(i=0; i<DIM(mode_f81439_item, MODE_F81439_INFO); i++)
	{
		info = &mode_f81439_item[i];

		if(info->cmd_gpio == cmd_gpio)
			break;
	}

	return info->mode_str;

} /* END: find_mode_by_cmd_gpio() */

/*---------------------------------------------------------------------------------*/
/**
 * @return
 * 0: success
 * 1: get_transceiver_type failed
 * 2: get_current_mode failed
 */
int get_info(char *tty_port, TTY_ITEM *item, int *current_gpio_mode)
{
	int  io, idx, status, type, gpio_mode;
	int  ret			 = 0;
	char result_type[16] = {0};
	char result_mode[16] = {0};

	io  = read_ioport(tty_port);
	idx = get_tty_idx(tty_port, io);

	status = get_transceiver_type(tty_port, idx, &type);
	if(status)
	{
		sprintf(result_type, "Error(%d)", status);
		ret = 1;
		goto EXIT;
	}

	sprintf(result_type, "%s", transceiver_type_str[type]);

	status = get_current_mode(tty_port, idx, &gpio_mode);
	if(status)
	{
		sprintf(result_mode, "Error(%d)", status);
		ret = 2;
		goto EXIT;
	}

	if(gpio_mode == FIX_MODE)
		sprintf(result_mode, "Fix mode");
	else
		sprintf(result_mode, "%s", find_mode_str_by_cmd_gpio(gpio_mode));

EXIT:

	if(current_gpio_mode)
		*current_gpio_mode = gpio_mode;

	if(item)
	{
		item->list_info.io_addr = io;
		strcpy(item->tty_port, 		 tty_port);
		strcpy(item->list_info.type, result_type);
		strcpy(item->list_info.mode, result_mode);
	}

	return ret;

} /* END: get_info() */

/*---------------------------------------------------------------------------------*/
/**
 * @return
 * 0: success
 * 1: f81504_series module can not found
 * 2: tty can not found
 */
int list_port(void)
{
	int			   i, j;
    int            ret     		 = 0;
	int            port_cnt      = 0;
	TTY_ITEM	   item[CFG_MAX] = {0};
	TTY_ITEM	   item_tmp 	 = {0};
    DIR           *mod_d   		 = NULL;
	DIR           *dev_d   		 = NULL;
    struct dirent *mod_dir 		 = NULL;
    struct dirent *dev_dir 		 = NULL;
	char 		   dev_path[512] = {0};

    mod_d = opendir(MODULE_PATH);
    if(mod_d == NULL)
    {
        ret = 1;
        goto EXIT;
    }

    while((mod_dir = readdir(mod_d)) != NULL)
    {
		if(strstr(mod_dir->d_name, ":"))
		{
			memset(dev_path, 0, sizeof(dev_path));
			sprintf(dev_path, DEVICE_PATH, mod_dir->d_name);

			dev_d = opendir(dev_path);
			if(dev_d)
			{
				while((dev_dir = readdir(dev_d)) != NULL)
				{
					if(strstr(dev_dir->d_name, ".") == NULL)
					{
						if(get_info(dev_dir->d_name, &item[port_cnt], NULL) == 0)
							port_cnt++;

						if(port_cnt >= CFG_MAX)
							break;
					}
				}
				
				closedir(dev_d);
        		dev_d = NULL;

				if(port_cnt >= CFG_MAX)
					break;
			}
		}
    }

	if(port_cnt == 0)
		ret = 2;

	/* sorted by IO address */
	for(i=0; i<port_cnt; i++)
	{
		for(j=0; j<i; j++)
		{
			if(item[j].list_info.io_addr > item[i].list_info.io_addr)
			{
				memset(&item_tmp, 0, 		 sizeof(TTY_ITEM));
				memcpy(&item_tmp, &item[j],  sizeof(TTY_ITEM));
				memcpy(&item[j],  &item[i],  sizeof(TTY_ITEM));
				memcpy(&item[i],  &item_tmp, sizeof(TTY_ITEM));
			}
		}
	}

	printf(LIST_FORMAT, "Address", "Name", "Transceiver", "Mode");
	for(i=0; i<port_cnt; i++)
	{
		char io_addr_str[16] = {0};

		sprintf(io_addr_str, "%X", item[i].list_info.io_addr);

		printf(LIST_FORMAT, io_addr_str,
							item[i].tty_port,
							item[i].list_info.type,
							item[i].list_info.mode);
	}

EXIT:

    if(dev_d)
    {
        closedir(dev_d);
        dev_d = NULL;
    }

    if(mod_d)
    {
        closedir(mod_d);
        mod_d = NULL;
    }

	return ret;

} /* END: list_port() */

/*---------------------------------------------------------------------------------*/
/**
 * @return
 * 0: success
 * 1: open device config failed
 * 2: change mode failed
 */
int set_mode(char *tty_port, int mode)
{
	int  status;
	int  ret 	  = 0;
	int  fd 	  = -1;
	char buf[128] = {0};

	sprintf(buf, F81504_CONFIG_SYSFS, tty_port);

	fd = open(buf, O_RDWR);
	if(fd < 0)
	{
		ret = 1;
		goto EXIT;
	}

	status = change_mode(fd, mode, read_ioport(tty_port));
	if(status)
	{
		printf("change_mode failed !! (status: %d)\n", status);
		ret = 2;
		goto EXIT;
	}

EXIT:

	if(fd != -1)
	{
		close(fd);
		fd = -1;
	}

	return ret;

} /* END: set_mode() */

/*---------------------------------------------------------------------------------*/
/**
 * @return
 * 0: success
 * 1: change gpio failed
 */
int set_gpio(char *tty_port, int gpio)
{
	int  status, io, idx;
	int  ret 	  = 0;

	io  = read_ioport(tty_port);
	idx = get_tty_idx(tty_port, io);

	status = change_gpio(tty_port, idx, gpio);
	if(status)
	{
		printf("change_gpio failed !! (status: %d)\n", status);
		ret = 1;
		goto EXIT;
	}

EXIT:

	return ret;

} /* END: set_gpio() */
