#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <unistd.h>
#include <fcntl.h>
#include <getopt.h>
#include <errno.h>

#include "version.h"
#include "f81504.h"
#include "parson.h"

/*---------------------------------------------------------------------------------*/
#define DEFAULT_CFG "/etc/serial_f81504.cfg"

/*---------------------------------------------------------------------------------*/
typedef struct
{
    char config[256];
    char tty_port[16];
    char mode[8];
    bool save;
    bool init;
    bool list;
    bool reset;
    bool option;

} OPTION_INFO;

/*---------------------------------------------------------------------------------*/
static const struct option long_opt[] =
{
    { "help",        1, NULL, 'h' },
    { "config",      1, NULL, 'c' },
    { "ttyPort",     1, NULL, 't' },
    { "mode",        1, NULL, 'm' },
    { "withoutSave", 0, NULL, 'w' },
    { "init",        0, NULL, 'i' },
    { "list",        0, NULL, 'l' },
    { "reset",       0, NULL, 'r' },
    { "option",      0, NULL, 'o' },
    {  NULL,         0, NULL, '0' },
};

static OPTION_INFO opt_info =
{
    .config     = DEFAULT_CFG,
    .tty_port   = {0},
    .mode       = {0},
    .save       = true,
    .init       = false,
    .list       = false,
    .reset      = false,
    .option     = false,
};

static int       tty_item_cnt = 0;
static TTY_ITEM *tty_item     = NULL;

/*---------------------------------------------------------------------------------*/
static void show_usage (char *name)
{
    char *ttyPort_sample = "ttyS4";

    printf("Syntax: %s [-option ...]                                \n", name);
    printf("Option:                                                 \n");
    printf("  -h : Show usage                                       \n");
    printf("  -c : Target of config path (default: %s)              \n", DEFAULT_CFG);
    printf("  -t : tty port (ex: %s)                                \n", ttyPort_sample);
    printf("  -m : tty mode (ex: RS422 or 0)                        \n");
    printf("       Mode table (M0/M1/M2):                           \n");
    printf("        0: 000 - RS422                                  \n");
    printf("        1: 001 - RS232                                  \n");
    printf("        2: 010 - RS485                                  \n");
    printf("        4: 100 - RS422R (with termination)              \n");
    printf("        6: 110 - RS485R (with termination)              \n");
    printf("  -w : Without save config                              \n");
    printf("  -i : Initialize device by config                      \n");
    printf("  -l : Display all tty port and exit                    \n");
    printf("  -r : Reset config to default                          \n");
    printf("  -o : show option for debug                            \n");
    printf("Sample:                                                 \n");
    printf("  Change tty's mode:                                    \n");
    printf("       %s -t %s -m RS422                                \n", name, ttyPort_sample);
    printf("  Change tty's mode without save config:                \n");
    printf("       %s -t %s -m RS422 -w                             \n", name, ttyPort_sample);
    printf("  Initialize device by config (default: %s):            \n", DEFAULT_CFG);
    printf("       %s -i                                            \n", name);
    printf("  Initialize device by target config                    \n");
    printf("       %s -i -c ./init.cfg                              \n", name);
    printf("  Display all tty port:                                 \n");
    printf("       %s -l                                            \n", name);
    printf("  Reset config to default:                              \n");
    printf("       %s -r                                            \n", name);

} /* END: show_usage() */

/*---------------------------------------------------------------------------------*/
static void show_option (void)
{
    printf("Option\n");
    printf("--------------------------------------------------------\n");
    printf("config     : %s\n", opt_info.config);
    printf("tty_port   : %s\n", opt_info.tty_port);
    printf("mode       : %s\n", opt_info.mode);
    printf("save       : %s\n", opt_info.save   ? "True" : "False");
    printf("init       : %s\n", opt_info.init   ? "True" : "False");
    printf("list       : %s\n", opt_info.list   ? "True" : "False");
    printf("reset      : %s\n", opt_info.reset  ? "True" : "False");
    printf("option     : %s\n", opt_info.option ? "True" : "False");
    printf("--------------------------------------------------------\n");

} /* END: show_option() */

/*---------------------------------------------------------------------------------*/
static MODE_F81439_INFO *get_mode_f81439_info (char *mode)
{
	int				  i;
	MODE_F81439_INFO *info = NULL;

	for(i=0; i<DIM(mode_f81439_item, MODE_F81439_INFO); i++)
	{
		info = &mode_f81439_item[i];

		if((strcmp(info->mode_str, mode) == 0) || (strcmp(info->mode_int, mode) == 0))
			break;
	}

	return info;

} /* END: get_mode_f81439_info() */

/*---------------------------------------------------------------------------------*/
static int change_interface (char *tty_port, char *mode, MODE_F81439_INFO *extern_mode_info)
{
    int               ret = 0;
    int               gpio_mode;
    MODE_F81439_INFO *mode_info = NULL;

    if(is_f81504(tty_port))
    {
        printf("Not F81504 series !!\n");
        ret = 1;
        goto EXIT;
    }

    ret = get_info(tty_port, NULL, &gpio_mode);
    if(ret)
    {
        printf("serial_fix_mode !! (ret: %d)\n", ret);
        ret = 2;
        goto EXIT;        
    }

    if(gpio_mode == FIX_MODE)
    {
        printf("%s is gpio_fix_mode !!\n", tty_port);
        ret = 3;
        goto EXIT;
    }

    if(extern_mode_info)
        mode_info = extern_mode_info;
    else
        mode_info = get_mode_f81439_info(mode);

    if(mode_info->cmd_gpio == -1)
    {
        printf("Invalid mode !!\n");
        ret = 4;
        goto EXIT;
    }

    if(mode_info->cmd_mode != -1)
    {
        ret = set_mode(tty_port, mode_info->cmd_mode);
        if(ret)
        {
            printf("set_mode failed !! (ret: %d)\n", ret);
            ret = 5;
            goto EXIT;
        }
    }

    if(mode_info->cmd_gpio != -1)
    {
        ret = set_gpio(tty_port, mode_info->cmd_gpio);
        if(ret)
        {
            printf("set_gpio failed !! (ret: %d)\n", ret);
            ret = 6;
            goto EXIT;
        }
    }

    printf("%s already set mode to %s !!\n", tty_port, mode_info->mode_str);

EXIT:

    return ret;

} /* END: change_interface() */

/*---------------------------------------------------------------------------------*/
static TTY_ITEM *find_item_by_port (char *tty_port)
{
    int       i;
    TTY_ITEM *item = NULL;

    for(i=0; i<tty_item_cnt; i++)
    {
        item = &tty_item[i];

        if(strcmp(item->tty_port, tty_port) == 0)
            return item;
    }

    return NULL;

} /* END: find_item_by_port() */

/*---------------------------------------------------------------------------------*/
static void append_tty_item (char *tty_port, MODE_F81439_INFO *info)
{
    TTY_ITEM *item    = NULL;
    TTY_ITEM *tmp_ptr = NULL;

    if(tty_item)
    {
        item = find_item_by_port(tty_port);

        if(item)
            item->mode = info;
        else
        {
            tmp_ptr = realloc(tty_item, (sizeof(TTY_ITEM) * (tty_item_cnt + 1)));
            if(tmp_ptr == NULL)
            {
                printf("realloc fail !!");
                return;
            }
            tty_item = tmp_ptr;

            memset(&tty_item[tty_item_cnt], 0, sizeof(TTY_ITEM));
            strcpy(tty_item[tty_item_cnt].tty_port, tty_port);
            tty_item[tty_item_cnt].mode = info;

            tty_item_cnt++;
        }
    }
    else
    {
        tty_item = malloc(sizeof(TTY_ITEM));
        memset(&tty_item[0], 0, sizeof(TTY_ITEM));
        strcpy(tty_item[0].tty_port, tty_port);
        tty_item[0].mode = info;

        tty_item_cnt++;
    }


} /* END: append_tty_item() */

/*---------------------------------------------------------------------------------*/
static int json_2_cfg (const char *buf)
{
    int               i;
    int               ret  = 0;
    JSON_Value       *val  = NULL;
    JSON_Object      *obj  = NULL;
    const char       *port = NULL;
    const char       *mode = NULL;
    MODE_F81439_INFO *info = NULL;
    char              key[16];

    val = json_parse_string(buf);
    if(val == NULL)
    {
        ret = 1;
        printf("JSON_Value fail !!\n");
        goto EXIT;
    }

    obj = json_value_get_object(val);
    if(obj == NULL)
    {
        ret = 2;
        printf("JSON_Object fail !!\n");
        goto EXIT;
    }

    for(i=0; i<CFG_MAX; i++)
    {
        memset(key, 0, sizeof(key));
        sprintf(key, CFG_PORT_FORMAT, i);
        port = json_object_get_string(obj, key);

        memset(key, 0, sizeof(key));
        sprintf(key, CFG_MODE_FORMAT, i);
        mode = json_object_get_string(obj, key);

        if(port && mode)
        {
            info = get_mode_f81439_info((char*) mode);
            if(info->cmd_gpio != -1)
                append_tty_item((char*) port, info);
        }

        port = NULL;
        mode = NULL;
        info = NULL;
    }

EXIT:

    if(val)
        json_value_free(val);

    return ret;

} /* END: json_2_cfg() */

/*---------------------------------------------------------------------------------*/
static int cfg_2_json (char *buf)
{
    int          i;
    int          ret        = 0;
    JSON_Value  *val        = NULL;
    JSON_Object *obj        = NULL;
    char        *serialized = NULL;
    char         key[16]    = {0};

    val = json_value_init_object();
    obj = json_value_get_object(val);    

    for(i=0; i<tty_item_cnt; i++)
    {
        memset(key, 0, sizeof(key));
        sprintf(key, CFG_PORT_FORMAT, i);
        json_object_set_string(obj, key, tty_item[i].tty_port);

        memset(key, 0, sizeof(key));
        sprintf(key, CFG_MODE_FORMAT, i);
        json_object_set_string(obj, key, tty_item[i].mode->mode_str);
    }

    serialized = json_serialize_to_string(val);
    memcpy(buf, serialized, json_serialization_size(val));

    if(serialized)
        json_free_serialized_string(serialized);

    if(val)
        json_value_free(val);

    return ret;

} /* END: cfg_2_json() */

/*---------------------------------------------------------------------------------*/
static int export_cfg (char *file)
{
    int   ret       = 0;
    FILE *fp        = NULL;
    char  buf[2048] = {0};

    cfg_2_json(buf);

    fp = fopen(file, "w");

    if(fp == NULL)
    {
        ret = 1; 
        printf("Open %s fail !! error = %d, %s !!\n", file, errno, strerror(errno));
        goto EXIT;        
    }

    fprintf(fp, "%s", buf);

EXIT:

    if(fp)
    {
        fclose(fp);
        sync();
    }

    return ret;


} /* END: export_cfg() */

/*---------------------------------------------------------------------------------*/
static int import_cfg (char *file)
{
    int   len;
    int   ret = 0;
    FILE *fp  = NULL;
    char *buf = {0};

    fp = fopen(file, "r");
    if(fp == NULL)
    {
        printf("Open %s fail !! error = %d, %s !!\n", file, errno, strerror(errno));
        ret = 1;
        goto EXIT;        
    }

    fseek(fp, 0, SEEK_END);
    len = ftell(fp);
    fseek(fp, 0, SEEK_SET);

    buf = (char *) malloc(len);
    fread(buf, len, 1, fp);

    ret = json_2_cfg(buf);
    if(ret)
    {
        printf("json_2_cfg fail !! (ret: %d)\n", ret);
        ret = 2; 
        goto EXIT;        
    }

EXIT:

    if(buf)
        free(buf);

    if(fp)
        fclose(fp);

    return ret;

} /* END: import_cfg() */

/*---------------------------------------------------------------------------------*/
static void save_to_cfg (char *config, char *tty_port, char *mode)
{
    MODE_F81439_INFO *info = NULL;

    info = get_mode_f81439_info(mode);
    if(info->cmd_gpio != -1)
    {
        if(access(config, F_OK) == 0)
            import_cfg(config);

        append_tty_item(tty_port, info);

        export_cfg(config);
    }

} /* END: save_to_cfg() */

/*---------------------------------------------------------------------------------*/
static void init_by_cfg (char *exec_name, char *config)
{
    int       i;
    int       ret  = 0;
    TTY_ITEM *item = NULL;

    if(access(config, F_OK) == 0)
        import_cfg(config);
    else
    {
        printf("No config. (default RS232)\n");
        return;
    }

    for(i=0; i<tty_item_cnt; i++)
    {
        item = &tty_item[i];

        ret = change_interface(item->tty_port, NULL, item->mode);
        if(ret)
            printf("change_interface failed !! (%s, ret: %d)\n", item->tty_port, ret);
    }

} /* END: init_by_cfg() */

/*---------------------------------------------------------------------------------*/
int main (int argc, char **argv)
{
    int ret, opt;

    printf("version: %s\n", SERIAL_F81504_TOOL_VERSION);

    if(argc < 2)
    {
        show_usage(argv[0]);
        goto EXIT;
    }

    while((opt = getopt_long(argc, argv, "?hc:t:m:wilro", long_opt, NULL)) != -1)
    {
        switch(opt)
        {
            case 'c':
                        memset(opt_info.config, 0, sizeof(opt_info.config));
                        memcpy(opt_info.config,   optarg, strlen(optarg));
                        break;
            case 't':
                        memcpy(opt_info.tty_port, optarg, strlen(optarg));
                        break;
            case 'm':
                        memcpy(opt_info.mode,     optarg, strlen(optarg));
                        break;
            case 'w':
                        opt_info.save   = false;
                        break;
            case 'i':
                        opt_info.init   = true;
                        break;
            case 'l':
                        opt_info.list   = true;
                        break;
            case 'r':
                        opt_info.reset  = true;
                        break;
            case 'o':
                        opt_info.option = true;
                        break;
            case 'h':
            case '?':
            default:
                        show_usage(argv[0]);
                        goto EXIT;
        }
    }

    if(opt_info.option)
        show_option();

    if(opt_info.list)
    {
        ret = list_port();
        if(ret)
            printf("f81504_series module can not found !! (ret: %d)\n", ret);
        goto EXIT;
    }

    if(opt_info.reset)
    {
        printf("Remove \"%s\"\n", opt_info.config);
        remove(opt_info.config);
        goto EXIT;
    }

    if(opt_info.init)
    {
        init_by_cfg(argv[0], opt_info.config);
        goto EXIT;
    }

    if((strlen(opt_info.tty_port) == 0) || (strlen(opt_info.mode) == 0))
    {
        show_usage(argv[0]);
        goto EXIT;
    }

    ret = change_interface(opt_info.tty_port, opt_info.mode, NULL);
    if(ret)
    {
        printf("change_interface failed !! (ret: %d)\n", ret);
        goto EXIT;
    }

    if(opt_info.save)
        save_to_cfg(opt_info.config, opt_info.tty_port, opt_info.mode);

EXIT:

    if(tty_item)
        free(tty_item);

    return 0;

} /* END: main() */
